﻿/*
 * 
 *          DremDOS
 *          VGA Terminal Class (VGATerminal.cs)
 * 
 *          Contains the frontend part of Consoles. It's primary purpose is to be controlled by Terminal objects. It also has a graphical mode for graphical programs, if that's your thing.
 *
 */

using System;
using System.Collections.Generic;
using System.Text;
using Sys = Cosmos.System;
using Cosmos.HAL.Drivers.PCI.Video;
using Cosmos.System.Graphics;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Cosmos.System;
using System.Drawing;
using Cosmos.Core.IOGroup;
using Console = System.Console;
using DremDOS.Core.Graphics;
using DremDOS;

namespace DremDOS.Core.Terminal
{
    class VGATerminal
    {
        public bool ShowCursor = false;
        public bool EnableWrite = false;
        public bool UseTerminal = true;
        public string VGATerminalName = "Shh... let's not leak our hard work!";
        public Pen BackgroundColor = new Pen(Color.Black);
        public Pen ForegroundColor = new Pen(Color.White);
        public bool CanBackspace = true;

        private int PositionX;
        private int PositionY;
        private int TextWidth;
        private int TextHeight;
        private int CursorX = 0;
        private int CursorY = 0;
        private Cosmos.System.Graphics.Point PointXY;
        private char[][] TextBuffer;
        private Pen[][] ColorBuffer;
        private bool Hidden = false;
        private Terminal LinkedTerminal;
        private Sys.Graphics.Fonts.PCScreenFont Font = Sys.Graphics.Fonts.PCScreenFont.Default;

        private Canvas canvas;
        //private Canvas ConsoleCanvas;

        private Pen WhitePen = new Pen(Color.White);
        private Pen BlackPen = new Pen(Color.Black);
        public VGATerminal(Terminal Term, int TextWidth, int TextHeight, int PositionX, int PositionY, Kernel kernel)
        {
            canvas = kernel.GetCanvas();
            Initialize(Term, TextWidth, TextHeight, PositionX, PositionY);
        }
        /*public VGATerminal(Terminal term, int TextWidth, int TextHeight, int PositionX, int PositionY)
        {
            Initialize(term, TextWidth, TextHeight, PositionX, PositionY);
        }*/

        private void Initialize(Terminal Term, int Width, int Height, int X, int Y)
        {
            PositionX = X;
            PositionY = Y;
            PointXY = new Cosmos.System.Graphics.Point(X, Y);
            TextWidth = Width;
            TextHeight = Height;
            LinkedTerminal = Term;

            // Initialize text buffer
            TextBuffer = new char[Height][];
            for (int i = 0; i<Height; i++) {
                TextBuffer[i] = new char[Width];
                for(int j = 0; j<Width; j++)
                {
                    TextBuffer[i][j] = ' ';
                }
            }
            // Initialize color buffer
            ColorBuffer = new Pen[Height][];
            for (int i = 0; i < Height; i++)
            {
                ColorBuffer[i] = new Pen[Width];
                for (int j = 0; j < Width; j++)
                {
                    ColorBuffer[i][j] = ForegroundColor;
                }
            }
        }
        private void MoveConsole(Canvas canvas, int PositionX, int PositionY)
        {

        }
        public void WriteText(string Text, int x, int y) // Writes text to the TextBuffer. Sets associated ColorBuffer to white color (default).
        {
            //CoverCursor();
            char[] Temp = Text.ToCharArray();
            for (int i = 0; i < Temp.Length; i++)
            {
                if (Temp[i] == '\n') { NextLine(); y = CursorY; x = CursorX; }
                else
                {
                    if (x > (TextWidth-1)) { NextLine(); x = CursorX; y = CursorY; }
                    TextBuffer[y][x] = Temp[i];
                    ColorBuffer[y][x] = ForegroundColor;
                    if (!Hidden) UpdateCharacter(x, y);
                    x++;
                }
            }
            SetCursorPosition(x, y);
            if(ShowCursor) { WriteCursor(); }
        }
        public void WriteText(string Text, Pen Color, int x, int y) // Writes text to the TextBuffer. Lets the programmer set the string's color.
        {
            if (ShowCursor) { CoverCursor(); }
            char[] Temp = Text.ToCharArray();
            for (int i = 0; i < Text.Length; i++)
            {
                if (Temp[i] == '\n') { NextLine(); y = CursorY; x = CursorX; }
                else
                {
                    if (x > TextWidth - 1) { NextLine(); x = CursorX; y = CursorY; }
                    TextBuffer[y][x] = Temp[i];
                    ColorBuffer[y][x] = Color;
                    if (!Hidden) UpdateCharacter(x, y);
                    x++;
                }
            }
            SetCursorPosition(x, y);
            if (ShowCursor) { WriteCursor(); }
        }
        public void WriteText(string Text, Pen[] Color, int x, int y) // Writes text to the TextBuffer. Lets the programmer set each character's color.
        {
            if (ShowCursor) { CoverCursor(); }
            char[] Temp = Text.ToCharArray();
            for (int i = 0; i < Text.Length; i++)
            {
                if (Temp[i] == '\n') { NextLine(); y = CursorY; x = CursorX; }
                else
                {
                    if (x > TextWidth - 1) { NextLine(); x = CursorX; y = CursorY; }
                    TextBuffer[y][x] = Temp[i];
                    ColorBuffer[y][x] = Color[i];
                    if (!Hidden) UpdateCharacter(x, y);
                    x++;
                }
            }
            SetCursorPosition(x, y);
            if (ShowCursor) { WriteCursor(); }
        }
        public void AppendText(string Text) // Writes text to where the cursor is.
        {
            WriteText(Text, CursorX, CursorY);
        }
        public void AppendText(char Character) // Writes text to where the cursor is.
        {
            WriteText(Character.ToString(), CursorX, CursorY);
        }
        public void AppendText(string Text, Pen Color) // Writes text to where the cursor is, with a custom color.
        {
            WriteText(Text, Color, CursorX, CursorY);
        }
        public void InterpretKeyInput(ConsoleKeyInfo Character) // Writes text to where the cursor is.
        {
            if (Character.KeyChar == '\n') {
                CoverCursor();
                NextLine();
                WriteCursor();
                if (UseTerminal) LinkedTerminal.EnterPressed();
            } else if (Character.KeyChar == '\b')
            {
                //if (CursorX == 0 && CursorY != 0) { SetCursorPosition(TextWidth-1, CursorY-1); } else { if (CursorX != 0) SetCursorPosition(CursorX-1, CursorY); }
                if (CanBackspace)
                {
                    if (CursorX == 0 && CursorY != 0) { LastLine(); return; } else { if (CursorX != 0) SetCursorPosition(CursorX - 1, CursorY); }
                    TextBuffer[CursorY][CursorX] = ' ';
                    canvas.DrawFilledRectangle(BackgroundColor, new Sys.Graphics.Point(PositionX + (CursorX * 8), PositionY + (CursorY * 16)), 8, 16);
                    UpdateCharacter(CursorX, CursorY);
                    WriteCursor();
                }
                if (UseTerminal) LinkedTerminal.BufferBackspace();
            } else {
                WriteText(Character.KeyChar.ToString(), CursorX, CursorY);
                if (UseTerminal) SendToTerminal(Character.KeyChar);
            }
        }
        public void SendToTerminal(char Send)
        {
            LinkedTerminal.WriteToBuffer(Send);
        }
        public void SendToTerminal(string Send)
        {
            char[] temp = Send.ToCharArray();
            for(int i = 0; i < temp.Length; i++)
                LinkedTerminal.WriteToBuffer(temp[i]);
        }
        public void RunTerminalCommand()
        {
            LinkedTerminal.RunCommand();
        }
        public void RunTerminalCommand(string Run)
        {
            LinkedTerminal.RunCommand(Run);
        }
        public void NextLine() // Makes the cursor go to the next line.
        {
            CursorX = 0;
            if(CursorY >= TextHeight-1) { ScrollText(); } else { CursorY++; }
        }
        public void LastLine()
        {
            if(CursorY != 0)
            {
                for(int i = TextWidth-1; i >= 0; i--)
                {
                    if(TextBuffer[CursorY-1][i] != '\0')
                    {
                        SetCursorPosition(i+1, CursorY - 1);
                        return;
                    }
                }
                SetCursorPosition(0, CursorY - 1);
            }
        }
        public void ScrollText() // Scrolls text.
        {
            for(int i = 0; i < TextHeight-1; i++) { TextBuffer[i] = TextBuffer[i + 1]; }
            TextBuffer[TextHeight-1] = new char[TextWidth];
            for (int i = 0; i < TextWidth; i++) { TextBuffer[TextHeight - 1][i] = ' ';  }
            
            for (int i = 0; i < TextHeight - 1; i++) { ColorBuffer[i] = ColorBuffer[i + 1]; }
            ColorBuffer[TextHeight - 1] = new Pen[TextWidth];
            for (int i = 0; i < TextWidth; i++) { ColorBuffer[TextHeight-1][i] = WhitePen; }

            UpdateEntireScreen(true);
        }
        private void CoverCursor() // Draws a rectangle over the cursor.
        {
            if (ShowCursor)
            {
                if (CursorX != TextWidth)
                {
                    canvas.DrawFilledRectangle(BackgroundColor, new Sys.Graphics.Point(PositionX + (CursorX * 8), (PositionY + ((CursorY + 1) * 16)) - 1), 7, 1);
                }
                else if (CursorY == TextHeight - 1)
                {
                    return;
                }
                else
                {
                    canvas.DrawFilledRectangle(BackgroundColor, new Sys.Graphics.Point(PositionX, (PositionY + ((CursorY + 2) * 16)) - 1), 7, 1);
                }
            }
        }
        private void WriteCursor() // Draws a white rectangle where the cursor is.
        {
            if (ShowCursor) {
                if (CursorX != TextWidth)
                {
                    canvas.DrawFilledRectangle(ForegroundColor, new Sys.Graphics.Point(PositionX + (CursorX * 8), (PositionY + ((CursorY + 1) * 16)) - 1), 7, 1);
                } else if (CursorY == TextHeight - 1)
                {
                    return;
                } else
                {
                    canvas.DrawFilledRectangle(ForegroundColor, new Sys.Graphics.Point(PositionX, (PositionY + ((CursorY + 2) * 16)) - 1), 7, 1);
                }
            }
        }
        public void UpdateEntireScreen(bool DoBlank) // This takes a VERY long time. It's recommended to make good use of the UpdateLine methods instead.
        {
            if (DoBlank) { canvas.DrawFilledRectangle(BackgroundColor, new Sys.Graphics.Point(PositionX, PositionY), TextWidth * 8, TextHeight * 16); }

            //string Temp;
            for(int i = 0; i<TextHeight; i++)
            {
                //Temp = "";
                /*for (int j = 0; j<TextWidth; j++)
                {
                    //Temp += TextBuffer[i][j];
                    canvas.DrawACSIIString(ColorBuffer[i][j], TextBuffer[i][j].ToString(), PositionX + (8 * j), PositionY + (16 * i));
                    //canvas.DrawString(TextBuffer[i][j].ToString(), Font, ColorBuffer[i][j], PositionX + (8 * j), PositionY + (16 * i));
                    //Console.Beep(1500, 50);
                }*/
                //canvas.DrawACSIIString(ColorBuffer[i][j], Temp, PositionX, PositionY + (16*i));
                UpdateLine(i);
            }
            //CursorX = TextWidth;
            //CursorY = TextHeight;
            //CursorX = 0;
            //CursorY = 0;
            if (ShowCursor) { WriteCursor(); }
        }
        
        public void UpdateLine(int Line) // Updates a line
        {
            for (int i = 0; i < TextWidth; i++)
            {
                //canvas.DrawACSIIString(ColorBuffer[Line][i], TextBuffer[Line][i].ToString(), PositionX + (8 * i), PositionY + (16 * Line));
                canvas.DrawString(TextBuffer[Line][i].ToString(), Font, ColorBuffer[Line][i], PositionX + (8 * i), PositionY + (16 * Line));
            }
            if(ShowCursor) { WriteCursor(); }
        }
        public void UpdateLine(int Line, int Start)
        {
            for (int i = Start; i < TextWidth; i++)
            {
                //canvas.DrawACSIIString(ColorBuffer[Line][i], TextBuffer[Line][i].ToString(), PositionX + (8 * i), PositionY + (16 * Line));
                canvas.DrawString(TextBuffer[Line][i].ToString(), Font, ColorBuffer[Line][i], PositionX + (8 * i), PositionY + (16 * Line));
            }
            //CursorX = TextWidth;
            //CursorY = Line;
            if (ShowCursor) { WriteCursor(); }
        }
        public void UpdateLine(int Line, int Start, int End)
        {
            for (int i = Start; i < End; i++)
            {
                //canvas.DrawACSIIString(ColorBuffer[Line][i], TextBuffer[Line][i].ToString(), PositionX + (8 * i), PositionY + (16 * Line));
                canvas.DrawString(TextBuffer[Line][i].ToString(), Font, ColorBuffer[Line][i], PositionX + (8 * i), PositionY + (16 * Line));
            }
            //CursorX = TextWidth;
            //CursorY = End;
            if (ShowCursor) { WriteCursor(); }
        }
        public void UpdateCharacter(int x, int y)
        {
            //canvas.DrawACSIIString(ColorBuffer[y][x], TextBuffer[y][x].ToString(), PositionX + (8 * x), PositionY + (16 * y));
            canvas.DrawString(TextBuffer[y][x].ToString(), Font, ColorBuffer[y][x], PositionX + (8 * x), PositionY + (16 * y));
        }

        public void SetCursorPosition(int PosX, int PosY)
        {
            if (ShowCursor) { CoverCursor(); }
            CursorX = PosX;
            CursorY = PosY;
            if (ShowCursor) { WriteCursor(); }
        }

        public void Hide() { Hidden = true; canvas.DrawFilledRectangle(BackgroundColor, new Sys.Graphics.Point(PositionX, PositionY), TextWidth * 8, TextHeight * 16); }
        public void Show() { Hidden = false; UpdateEntireScreen(true); }
        public bool IsHidden() { return Hidden; }
        public int GetCursorX() { return CursorX; }
        public int GetCursorY() { return CursorY; }
        public int GetTextWidth() { return TextWidth; }
        public int GetTextHeight() { return TextHeight; }
        public char ReadCharacter(int x, int y) { return TextBuffer[y][x]; }
    }
}
