﻿/*
 * 
 *          DremDOS
 *          Main Kernel Class (Kernel.cs)
 * 
 *          Contains the BUI (Border User Interface) and main controlling logic for Terminals.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Sys = Cosmos.System;
using System;
using System.Collections.Generic;
using System.Text;
using Cosmos.HAL.Drivers.PCI.Video;
using Cosmos.System.Graphics;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Cosmos.System;
using System.Drawing;
using Cosmos.Core.IOGroup;
using Console = System.Console;
using DremDOS.Core.Terminal;
using DremDOS.Core.Graphics;
using DremDOS.Core.Drivers;
using DremDOS.Core.Sound;
using Cosmos.Debug.Kernel;
using System.Reflection.Metadata;

namespace DremDOS
{
    public class Kernel : Sys.Kernel
    {
        public int _SCREEN_WIDTH = 800;
        public int _SCREEN_HEIGHT = 600;
        public string _OS_NAME = "DremDOS";
        public int _OS_VERSION_MAJOR = 0;
        public int _OS_VERSION_MINOR = 1;
        public int _OS_VERSION_PATCH = 0;
        public string _OS_VERSION_TACKON = " (Multitasking Console Test)";
        public string _OS_VERSION_FULL = "Unknown";
        public int CurrentConsole = 1;
        public int TotalConsoles = 1;
        public Color BorderColor = Color.Purple;
        public Pen BorderPen = new Pen(Color.Purple);
        public Pen WhitePen = new Pen(Color.White);

        private Canvas canvas;
        private Terminal MainTerminal;
        //private static VGATerminal ConsoleA = new VGATerminal(MainTerminal, 80, 25, 50, 50);
        private VGATerminal ConsoleA;
        public MouseDriver mouseDriver;
        public Sys.Graphics.Fonts.PCScreenFont Font;
        //VMWareSVGAII driver;

        private bool ConsoleMode;
        private bool BlackBoxHasBeenFixed;

        private VGATerminal SelectedConsole;

        protected override void BeforeRun()
        {
            // Temporary stuff to get this thing to BEHAVE
            _OS_VERSION_FULL = _OS_VERSION_MAJOR + "." + _OS_VERSION_MINOR + "." + _OS_VERSION_PATCH + _OS_VERSION_TACKON;

            mDebugger.Send("BeforeRun start");
            ConsoleMode = false;
            Console.Clear();
            Sound.Startup();
            Console.ForegroundColor = ConsoleColor.Blue;
            Console.WriteLine(@"                                           _..._");
            Console.ForegroundColor = ConsoleColor.Red;
            Console.Write(@"    ___                ___  ____  ____   ");
            Console.ForegroundColor = ConsoleColor.DarkBlue;
            Console.WriteLine(@".:     '.");
            Sound.Startup();
            Console.ForegroundColor = ConsoleColor.Yellow;
            Console.Write(@"   / _ \_______ __ _  / _ \/ __ \/ __/  ");
            Console.ForegroundColor = ConsoleColor.DarkBlue;
            Console.WriteLine(@":::::     :");
            Sound.Startup();
            Console.ForegroundColor = ConsoleColor.Green;
            Console.Write(@"  / // / __/ -_)  ' \/ // / /_/ /\ \    ");
            Console.ForegroundColor = ConsoleColor.DarkMagenta;
            Console.WriteLine(@":::::::   :");
            Sound.Startup();
            Console.ForegroundColor = ConsoleColor.Blue;
            Console.Write(@" /____/_/  \__/_/_/_/____/\____/___/    ");
            Console.ForegroundColor = ConsoleColor.DarkMagenta;
            Console.WriteLine(@"`::::::::.'");
            Console.ForegroundColor = ConsoleColor.Magenta;
            Console.WriteLine(@"                                          `':::''");
            Console.ForegroundColor = ConsoleColor.Red;
            Console.WriteLine("\nNOTICE: THIS SOFTWARE IS PROVIDED AS-IS AND INCLUDES ABSOLUTELY NO WARRENTY. ANY DAMAGE TO YOUR DEVICE AND/OR DATA IS ON YOU!");
            Console.ForegroundColor = ConsoleColor.DarkRed;
            Console.WriteLine("YOU HAVE BEEN WARNED!");
            Console.ForegroundColor = ConsoleColor.Blue;
            Console.WriteLine("\nNote: Because of a bug in the PS/2 driver, Enter is *, Backspace is /");
            //Console.WriteLine("Left, Right, Up, and Down keys are Num4, Num6, Num8, and Num2, respectively.");
            Console.ForegroundColor = ConsoleColor.White;
            Console.WriteLine("\n");
            bool ValidResolutionChoiceFlag = false;
            while (!ValidResolutionChoiceFlag) {
                mDebugger.Send("Request resolution setting");
                Console.WriteLine("    1.) 800x600    2.) 1024x768    3.) 1280x720");
                Console.WriteLine("    4.) 1920x1080");
                Console.WriteLine("    C.) Single console mode");
                Console.WriteLine("Choose resolution:");
                string ResolutionChoice = Console.ReadKey().KeyChar.ToString();

                switch(ResolutionChoice) {
                    case "1":
                        _SCREEN_WIDTH = 800;
                        _SCREEN_HEIGHT = 600;
                        ValidResolutionChoiceFlag = true;
                        break;
                    case "2":
                        _SCREEN_WIDTH = 1024;
                        _SCREEN_HEIGHT = 768;
                        ValidResolutionChoiceFlag = true;
                        break;
                    case "3":
                        _SCREEN_WIDTH = 1280;
                        _SCREEN_HEIGHT = 720;
                        ValidResolutionChoiceFlag = true;
                        break;
                    case "4":
                        _SCREEN_WIDTH = 1920;
                        _SCREEN_HEIGHT = 1080;
                        ValidResolutionChoiceFlag = true;
                        break;
                    case "c":
                    case "C":
                        _SCREEN_WIDTH = 80;
                        _SCREEN_HEIGHT = 24;
                        ValidResolutionChoiceFlag = true;
                        ConsoleMode = true;
                        mDebugger.Send("Single console mode");
                        break;
                    default:
                        break;
                }
            }

            if (!ConsoleMode)
            {
                mDebugger.Send("Resolution has been set to " + _SCREEN_WIDTH + "x" + _SCREEN_HEIGHT);
                mDebugger.Send("Set font, init canvas and graphics driver");
                Font = Sys.Graphics.Fonts.PCScreenFont.Default;
                canvas = FullScreenCanvas.GetFullScreenCanvas(new Mode(_SCREEN_WIDTH, _SCREEN_HEIGHT, ColorDepth.ColorDepth32));
            }

        }

        protected override void Run()
        {   
            try
            {
                MainTerminal = new Terminal(ConsoleMode, this);
                if (!ConsoleMode)
                {
                    canvas.Clear(BorderColor);
                    canvas.Display();

                    mouseDriver = new MouseDriver(checked((uint)_SCREEN_WIDTH), checked((uint)_SCREEN_HEIGHT));

                    // TODO: Burn these magic numbers
                    ConsoleA = new VGATerminal(MainTerminal, (_SCREEN_WIDTH / 8) - 12, (_SCREEN_HEIGHT / 16) - 6, 50, 50, this);
                    ConsoleA.ShowCursor = true;
                    ConsoleA.EnableWrite = true;

                    MainTerminal.SetLinkedVGATerminal(ConsoleA);

                    canvas.Clear(BorderColor);
                    canvas.Display();

                    // TODO: Burn these magic numbers
                    canvas.DrawString(_OS_NAME + " " + _OS_VERSION_FULL, Font, WhitePen, 50, 5);
                    canvas.DrawString(ConsoleA.VGATerminalName, Font, WhitePen, 50, 20);

                    ConsoleA.UpdateEntireScreen(true);
                    
                    BUI.DrawBUI(canvas, this);
                    canvas.Display();
                    BlackBoxHasBeenFixed = false;
                    
                    MainTerminal.ShowOutput = true;
                    ConsoleA.RunTerminalCommand("command");

                    SelectedConsole = ConsoleA;
                } else
                {
                    // Yay sketti code cause past me wrote code like a mongol
                    mDebugger.Send("Setting up single console mode");
                    MainTerminal.ShowOutput = true;
                    MainTerminal.RunCommand("clear");
                    MainTerminal.RunCommand("command");
                }

                mDebugger.Send("Entering main OS loop.");
                while (true)
                {
                    //CurrentConsole = ConsoleA;
                    int InputCharInt = Console.Read();

                    if (InputCharInt != -1)
                    {
                        if (InputCharInt == 42) // Note: This will change to 8 once a bug with the PS/2 Controller is fixed
                        {
                            if (!ConsoleMode)
                            {
                                SelectedConsole.InterpretKeyInput(new ConsoleKeyInfo('\b', ConsoleKey.Backspace, false, false, false));
                            }
                            else
                            {
                                MainTerminal.BufferBackspace();
                            }
                        }
                        else if (InputCharInt == 47) // Some as the comment above except it'll be 13.
                        {
                            if (!ConsoleMode)
                            {
                                SelectedConsole.InterpretKeyInput(new ConsoleKeyInfo('\n', ConsoleKey.Enter, false, false, false));
                            }
                            else
                            {
                                MainTerminal.EnterPressed();
                            }
                        }
                        else
                        {
                            //ConsoleA.AppendText(((char)inputchar).ToString());
                            if (!ConsoleMode)
                            {
                                SelectedConsole.InterpretKeyInput(new ConsoleKeyInfo((char)InputCharInt, ConsoleKey.NoName, false, false, false));
                            }
                            else
                            {
                                MainTerminal.WriteToBuffer((char)InputCharInt);
                            }
                        }
                    }

                    if (!ConsoleMode)
                    {
                        mouseDriver.Draw(canvas);
                        canvas.Display();
                        
                        // Quick and dirty solution but dammit it works.
                        if (!BlackBoxHasBeenFixed && mouseDriver.HasMoved())
                        {
                            canvas.DrawFilledRectangle(BorderPen, new Sys.Graphics.Point(0, 0), 20, 20);
                            BlackBoxHasBeenFixed = true;
                            mDebugger.Send("Fixed black rectangle");
                        }
                    }
                }
            } catch (Exception e)
            {
                mDebugger.Send("EXCEPTION: " + e.GetType + ", " + e.Message);
                // Crash handler will eventually go here
            }
        }

        public Canvas GetCanvas() { return canvas; }
    }
}
